/*
--------------------------------------------------------------------------------

contra - a file (de-)compressor using symbol ranking by context similarity
(c) 2007, Frank Schwellinger, frank.schwellinger@gmx.de

state map + arithmetic encoder/decoder
(C) 2007, Matt Mahoney, matmahoney@yahoo.com

compiled with:
g++ -O2 -march=pentiumpro -Wall -fomit-frame-pointer -s -o contra contra-0.5.cpp

--------------------------------------------------------------------------------

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version. This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details at http://www.gnu.org/copyleft/gpl.html.

--------------------------------------------------------------------------------
HISTORY
--------------------------------------------------------------------------------
version  from date   until date   changes
-------  ----------- -----------  ----------------------------------------------
0.1      04-Dec-2007 14-Dec-2007  initial version
0.2      18-Dec-2007 21-Dec-2007  state map + arithmetic en-/decoder (taken from
                                  sr2) replaces former en-/decoder; added
                                  a ranktable based on longest common suffix
                                  successor symbol; secondary context builds
                                  from ordinary context plus successor symbol
0.3      22-Dec-2007 22-Dec-2007  allows 16/20/24 bit secondary contexts via
                                  commandline argument; small corrections
0.4      23-Dec-2007 23-Dec-2007  ranking context expanded to 12 bits; small
                                  corrections
0.5      24-Dec-2007 26-Dec-2007  adds binary search tree and secondary context
                                  construction mode; small modifications
--------------------------------------------------------------------------------
*/
//------------------------------------------------------------------------------
// INCLUDED HEADER FILES
//------------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
//------------------------------------------------------------------------------
// TYPE DEFINITIONS FOR INTEGERS
//------------------------------------------------------------------------------
typedef unsigned char UINT8;	//unsigned 8bit integer
typedef unsigned short UINT16;	//unsigned 16bit integer
typedef unsigned int UINT32;	//unsigned 32bit integer
typedef int SINT32;		//signed 32bit integer
//------------------------------------------------------------------------------
// GLOBALS
//------------------------------------------------------------------------------
//---Program Version
const UINT8 VERSION=0;
const UINT8 SUBVERSION=5;
//---Generated file header
const UINT8 MAGICBYTE=235; //identifies generated stream
UINT32 headerlength;
//---Program modes
UINT8 verbosemode;
UINT8 buffermode;
UINT8 suffixmode;
UINT8 tstmemmode;
UINT8 rebuildmode;
UINT8 lazymode;
UINT8 contextmode;
UINT8 ctxconstructionmode;
UINT8 searchmode;
//---other globals used by encoder and decoder
UINT32 ch=0;
UINT32 ctx=0;
UINT32 rkctx;
UINT32 pctx=0;
UINT32 mask=0;
UINT32 ctxmem=0;
UINT32 outbytes=0, inbytes=0;
double totaltime;
FILE *in,*out;
//---Ranking
const UINT32 NRANKS=256;
const UINT32 NRANKCTX=4096;
typedef struct ra
{
	UINT8 character;
	UINT16 count;
}ranktype;
ranktype ranktable[NRANKCTX][NRANKS];
//---input buffer
UINT32 inbufsize=0; 
UINT8 *inbuf; //input buffer (dynamic allocation at runtime)
UINT32 inbufprefix=258; //Size in bytes of input buffer prefix
UINT32 inbufpos=258; //(Start) position in the input buffer
//---Ternary search tree
typedef struct tnode *tstpointer;
typedef struct tnode {
	UINT16 splitchar;
	UINT32 matchpos;
	tstpointer lo, eq, hi;
	UINT8 visited;
}Tnode;
tstpointer root;
UINT32 tstmemsize=0;
UINT32 tstmaxnodes=0;
UINT32 tstnodecounter=0;
UINT8 *tstmemtest;
UINT32 predictedsymbol=0;
//---Suffix
UINT16 *suffix;
UINT32 sfxlen=0;
const UINT16 ENDOFSUFFIX=0x0100;
//---binary search tree
UINT32 *bstrightchild; //bst right child nodes (dynamic allocation at runtime)
UINT32 *bstleftchild; //bst left child nodes (dynamic allocation at runtime)
UINT32 bstnode=0;
UINT32 bstparent=0;
UINT32 bstcomparison=0;
UINT32 bstlimit=0;
UINT32 bstmatchlen=0;
UINT32 tmppredictedsymbol=0;

//------------------------------------------------------------------------------
// QUIT - Error handler: print message if any, and exit  - by Matt Mahoney
//------------------------------------------------------------------------------
void quit(const char* message=0)
{
	if (message) printf("%s\n", message);
	exit(1);
}

//------------------------------------------------------------------------------
// ALLOC  - Create an array p of n elements of type T - by Matt Mahoney
//------------------------------------------------------------------------------
template <class T> void alloc(T*&p, SINT32 n)
{
	p=(T*)calloc(n, sizeof(T));
	if (!p) quit("out of memory");
}

//------------------------------------------------------------------------------
// STATE MAP - maps a secondary context to a probability - by Matt Mahoney
//------------------------------------------------------------------------------
// Methods:
//
// Statemap sm(n)     creates a StateMap with n contexts using 4*n bytes memory.
// sm.p(cxt)          predicts next bit (0..4K-1) in context cxt (0..n-1).
// sm.update(cxt, y)  updates model for actual bit y (0..1) in cxt (0..n-1).
/*
A secondary context maps to a table entry containing a probability (p) that the
next bit will be a 1, and a count (n).  The entry is updated as follows:

  Input      (p, n) next state
  -------    -----------------
  Initial    (.5, 0)
    0        (p - p/(n+2), min(n+1, 127))
    1        (p + (1-p)/(n+2), min(n+1, 127))

p is represented as a 25 bit scaled unsigned integer, packed with n into 
a 32 bit integer (p in bits 8-31, n in 0-7).  To avoid division (which
is slow), the second term is computed by scaling p to 23 bits (discarding
2 low bits), multiplying by 512/(n+2) from a lookup table to produce
a 32 bit result, then scaling to 25 bits.
*/
//------------------------------------------------------------------------------
class StateMap
{
	protected:
		const SINT32 N;		// Number of contexts
		UINT32 *t;		// cxt -> prediction in high 25 bits,
					// count in low 7 bits
		static SINT32 dt[128];	// i -> 1K/(i+2)
	public:
		StateMap(SINT32 n);

		// predict next bit in context cxt
		SINT32 p(SINT32 cxt) {
			assert(cxt>=0 && cxt<N);
			return t[cxt]>>20;
		}

		// update model in context cxt for actual bit y
		void update(SINT32 cxt, SINT32 y) {
			assert(cxt>=0 && cxt<N);
			assert(y==0 || y==1);
			SINT32 n=t[cxt]&127, p=t[cxt]>>9; // count, prediction
			if (n<127) ++t[cxt];
			t[cxt]+=((y<<23)-p)*dt[n]&0xffffff80;
		}
};

SINT32 StateMap::dt[128]={0};

StateMap::StateMap(SINT32 n): N(n)
{
	alloc(t, N);
	for (SINT32 i=0; i<N; ++i)
		t[i]=1<<31;
	if (dt[0]==0)
		for (SINT32 i=0; i<128; ++i)
			dt[i]=512/(i+2);
}

//------------------------------------------------------------------------------
// ENCODER - does arithmetic encoding in n contexts - by Matt Mahoney
//------------------------------------------------------------------------------
// Methods:
//
// Encoder(f, n)  creates encoder for compression to archive f, which
//                must be open past any header for writing in binary mode.
// code(cxt, y)   compresses bit y (0 or 1) to file f
//                modeled in context cxt (0..n-1)
// flush()        should be called exactly once after compression is done and
//                before closing f.
/*
An arithmetic code represents a bit string S as a base-256 fraction between
x1 = p(<S) and x2 = p(<S)+p(S), where p(S) is the probability of S, and
p(<S) means the sum of the probabilities of all strings lexicographically
before S.  An arithmetic code can be computed incrementally because
p(S) = p(S1) * p(S2|S1) * p(S3|S1..2) * p(S4|S1..3) *...* p(Sn|S1..n-1)
where P(Si|S1..i-1) means the probability of the i'th bit of S given
the first i-1 bits.  Furthermore the range (x1,x2) (initially (0,1))
narrows given each new symbol probability.  As the leading base-256
digits of x1 and x2 match, they can be output.  We approximate (x1,x2)
so that only the first 4 nonmatching digits need to be kept in memory
(as 32 bit integers).

The arithmetic coder state is updated as follows.  Note that the
lexicographical order is reversed (1 before 0), which makes no difference.
The input is the bit to be coded (0 or 1) and the probability (p) from
the secondary context table.

  Input    (x1, x2) next state
  -------  -------------------
  Initial  (0, 2^32-1)
    0      ((x2-x1)*p+1, x2)
    1      (x1, (x2-x1)*p)

To avoid overflow, x2-x1 is scaled to 20 bits (discarding the low
12 bits) and p is scaled to 12 bits (discarding the low 13 bits).

During decompression, the bit is determined by comparing the compressed
data (x) with (x2-x1)*p.  If it is less or equal, then the next bit is
1, else 0.

x1 and x2 represent 4 digits of a base 256 number with the most
significant digit in bits 24-31.  If the leading
digits are equal, they are shifted out.  During decompression, the
next input byte would be shifted into x as well.

  while x1/256^3 = x2/256^3
    if compressing then output x1 / 256^3
    (x1, x2) <-- (x1 * 256 (mod 2^32), x2 * 256 + 255 (mod 2^32))
    if decompressing then x <-- x * 256 + next input byte (mod 2^32)

Thus, it is always true that x2 > x1, and there is never a case
where a bit cannot be coded.  In the worst case, x2 = x1 + 1, so
that (x2-x1)*p = x1 due to integer truncation and because 0 <= p < 1.
In this case (x1, x2) is updated to either (x1, x1) or (x2, x2), after
which 4 bytes are shifted and the new state is (0, 2^32-1).
*/
//------------------------------------------------------------------------------
class Encoder
{
	private:
		const SINT32 N;	// number of contexts
		FILE* archive;	// Compressed data file
		UINT32 x1, x2;	// Range, initially [0, 1), scaled by 2^32
		StateMap sm;	// cxt -> p
	public:
		Encoder(FILE* f, SINT32 n);
		void code(SINT32 cxt, SINT32 y);	// compress bit y
		void flush();			// call this when compression
						// is finished
};

// Compress bit y (0..1) in context cxt (0..n-1)
inline void Encoder::code(SINT32 cxt, SINT32 y)
{
	assert(y==0 || y==1);
	assert(cxt>=0 && cxt<N);
	SINT32 p=sm.p(cxt);
	assert(p>=0 && p<4096);
	UINT32 xmid=x1 + (x2-x1>>12)*p;
	assert(xmid>=x1 && xmid<x2);
	y ? (x2=xmid) : (x1=xmid+1);
	sm.update(cxt, y);
	while (((x1^x2)&0xff000000)==0) {	// pass equal leading bytes of
						// range
		putc(x2>>24, archive);
		outbytes++;
		x1<<=8;
		x2=(x2<<8)+255;
	}
}

Encoder::Encoder(FILE* f, SINT32 n):
	N(n), archive(f), x1(0), x2(0xffffffff), sm(n)
{
}

void Encoder::flush()
{
	putc(x1>>24, archive);	// Flush first unequal byte of range
}

//------------------------------------------------------------------------------
// DECODER - does arithmetic decoding in n contexts - by Matt Mahoney
//------------------------------------------------------------------------------
// Methods:
//
// Decoder(f, n)  creates decoder for decompression from archive f,
//                which must be open past any header for reading in binary mode.
// code(cxt)      returns the next decompressed bit from file f
//                with context cxt in 0..n-1.
//------------------------------------------------------------------------------
class Decoder
{
	private:
		const SINT32 N;	// number of contexts
		FILE* archive;	// Compressed data file
		UINT32 x1, x2;	// Range, initially [0, 1), scaled by 2^32
		UINT32 x;	// Decompress mode: last 4 input bytes of
				// archive
		StateMap sm;	// cxt -> p
	public:
		Decoder(FILE* f, SINT32 n);
		SINT32 code(SINT32 cxt);	// decompress a bit
};

// Return decompressed bit (0..1) in context cxt (0..n-1)
inline SINT32 Decoder::code(SINT32 cxt)
{
	assert(cxt>=0 && cxt<N);
	SINT32 p=sm.p(cxt);
	assert(p>=0 && p<4096);
	UINT32 xmid=x1 + (x2-x1>>12)*p;
	assert(xmid>=x1 && xmid<x2);
	SINT32 y=x<=xmid;
	y ? (x2=xmid) : (x1=xmid+1);
	sm.update(cxt, y);
	while (((x1^x2)&0xff000000)==0) {	// pass equal leading bytes of
						// range
		x1<<=8;
		x2=(x2<<8)+255;
		x=(x<<8)+(getc(archive)&255);	// EOF is OK
		inbytes++;
	}
	return y;
}

Decoder::Decoder(FILE* f, SINT32 n):
		N(n), archive(f), x1(0), x2(0xffffffff), x(0), sm(n)
{
	for (SINT32 i=0; i<4; ++i) {
		x=(x<<8)+(getc(archive)&255);
		inbytes++;
	}
}

//------------------------------------------------------------------------------
// TST GET ROOT
//------------------------------------------------------------------------------
tstpointer tstgetroot()
{
	tstpointer p;

	p=(tstpointer)malloc(sizeof(Tnode));
	p->splitchar=(UINT16)128;
	p->lo=p->eq=p->hi=0;
	p->visited=0;
	tstnodecounter=1;
	return p;
}

//------------------------------------------------------------------------------
// TST FREE NODES
//------------------------------------------------------------------------------
void tstfreenodes(tstpointer p)
{
	if(p){
		tstfreenodes(p->lo);
		if(p->eq){
			tstfreenodes(p->eq);
		}
		tstfreenodes(p->hi);
		free(p);
	}
}

//------------------------------------------------------------------------------
// TST CLEANUP
//------------------------------------------------------------------------------
void tstcleanup(tstpointer p)
{
	tstfreenodes(p);
	root=tstgetroot();
}

//------------------------------------------------------------------------------
// TST INSERT
//------------------------------------------------------------------------------
tstpointer tstinsert(tstpointer p,UINT32 s)
{
	if(p==0){
		tstnodecounter++;
		p=(tstpointer)malloc(sizeof(Tnode));
		p->splitchar=suffix[s];
		p->lo=p->eq=p->hi=0;
		p->visited=0;
	}
	p->matchpos=inbufpos;
	if(suffix[s]<p->splitchar){
		p->lo=tstinsert(p->lo,s);
	}else if(suffix[s]==p->splitchar){
		if(suffix[s]==ENDOFSUFFIX){
			//end of suffix
		}else{
			p->eq=tstinsert(p->eq,++s);
		}
	}else{
		p->hi=tstinsert(p->hi,s);
	}
	return p;
}

//------------------------------------------------------------------------------
// TST UPDATE
//------------------------------------------------------------------------------
SINT32 tstupdate(UINT32 s)
{
	tstpointer p,lastp=0;
	UINT32 matchlen=0;

	p=root;
	while(p){
		lastp=p;
		if(suffix[s]<p->splitchar){
			p=p->lo;
		}else if(suffix[s]==p->splitchar){
			predictedsymbol=inbuf[p->matchpos+1];
			matchlen++;
			p->matchpos=inbufpos;
			if(suffix[s++]==ENDOFSUFFIX){
				return 1;
			}
			p=p->eq;
		}else{
			p=p->hi;
		}
	}
	if(lazymode){ //lazy node creation
		if(lastp->visited>matchlen){
			//expand suffix tree only by 1 symbol
			suffix[s+1]=ENDOFSUFFIX;
			tstinsert(lastp,s);
		}
		if(lastp->visited<255)lastp->visited+=1;
	}else{
		//expand suffix tree only by 1 symbol
		suffix[s+1]=ENDOFSUFFIX;
		tstinsert(lastp,s);
	}

	return 0;
}

//------------------------------------------------------------------------------
// REBUILD TST
//------------------------------------------------------------------------------
inline void rebuildtst()
{
	UINT32 i=0,ii=0,currentinbufpos=0;

	if(verbosemode){
		printf("%d TST nodes created\n",tstnodecounter);
		printf("Rebuilding TST from last %d percent of input buffer\n",
			100-((10-rebuildmode)*10));
	}

	tstcleanup(root);
	currentinbufpos=inbufpos-inbufprefix;
	inbufpos=inbufprefix;

	for(i=((currentinbufpos/10)*(10-rebuildmode))+inbufprefix;
			i<currentinbufpos+inbufprefix+1;i++) {
		//---update input buffer
		inbufpos++;
		inbuf[inbufpos]=inbuf[i];
		//---insert suffix into TST
		for(ii=0;ii<sfxlen-1;ii++){
			suffix[ii]=(UINT16)inbuf[inbufpos-ii];
		}
		suffix[sfxlen-1]=ENDOFSUFFIX; //add end-of-suffix marker
		//---update tst
		tstupdate(0);
	}
}

//------------------------------------------------------------------------------
// UPDATE TST
//------------------------------------------------------------------------------
inline void updatetst(UINT8 symbol)
{
	UINT32 i=0;

	//---update input buffer
	inbufpos++;
	if(inbufpos==(UINT32)(inbufsize+inbufprefix)) {
		if(verbosemode){
			printf("Input buffer limit reached\n");
		}
		rebuildtst();
	}
	inbuf[inbufpos]=symbol;

	//---build suffix
	for(i=0;i<sfxlen-1;i++){
		suffix[i]=(UINT16)inbuf[inbufpos-i];
	}
	suffix[sfxlen-1]=ENDOFSUFFIX; //add end-of-suffix marker
	//---update tst
	tstupdate(0);
	if(tstnodecounter>=tstmaxnodes){
		if(verbosemode){
			printf("TST node limit reached\n");
		}
		rebuildtst();
	}
}

//------------------------------------------------------------------------------
// REBUILD BINARY SEARCH TREE
//------------------------------------------------------------------------------
inline void rebuildbst()
{
	UINT32 i=0,currentinbufpos=0;;

	if(verbosemode){
		printf("Rebuilding BST from last %d percent of input buffer\n",
			100-((10-rebuildmode)*10));
	}

	//---reset binary search tree child nodes
	for(i=0;i<inbufsize+inbufprefix;i++) {
		bstrightchild[i]=0;
		bstleftchild[i]=0;
	}

	//---partially rebuild binary search tree
	currentinbufpos=inbufpos-inbufprefix;
	inbufpos=inbufprefix;
	for(i=((currentinbufpos/10)*(10-rebuildmode))+inbufprefix;
			i<currentinbufpos+inbufprefix+1;i++) {
		//---update input buffer
		inbufpos++;
		inbuf[inbufpos]=inbuf[i];
		bstnode=inbufprefix;
		while(1) {
			if(bstnode==0) {
				bstcomparison=0;
				while(1) {
					if(inbuf[inbufpos-bstcomparison]>
					        inbuf[bstparent-bstcomparison]) {
						bstrightchild[bstparent]=inbufpos;
						break;
					}
					if(inbuf[inbufpos-bstcomparison]<
					        inbuf[bstparent-bstcomparison]) {
						bstleftchild[bstparent]=inbufpos;
						break;
					}
					bstcomparison++;
					if(bstcomparison>bstlimit) {
						break;
					}
				}
				break;
			} else {
				bstcomparison=0;
				while(1) {
					if(inbuf[inbufpos-bstcomparison]>
					        inbuf[bstnode-bstcomparison]) {
						bstparent=bstnode;
						bstnode=bstrightchild[bstnode];
						break;
					}
					if(inbuf[inbufpos-bstcomparison]<
					        inbuf[bstnode-bstcomparison]) {
						bstparent=bstnode;
						bstnode=bstleftchild[bstnode];
						break;
					}
					bstcomparison++;
					if(bstcomparison>bstlimit) {
						bstparent=bstnode;
						bstnode=0;
						break;
					}
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
// UPDATE BINARY SEARCH TREE
//------------------------------------------------------------------------------
inline void updatebst(UINT8 symbol)
{
	//---update input buffer
	inbufpos++;
	if(inbufpos==(UINT32)(inbufsize+inbufprefix)) {
		if(verbosemode){
			printf("Input buffer limit reached\n");
		}
		rebuildbst();
	}
	inbuf[inbufpos]=symbol;

	//---update binary search tree and get longest common suffix successor
	tmppredictedsymbol=0;
	predictedsymbol=0;
	bstmatchlen=0;
	bstnode=inbufprefix;
	while(1) {
		if(bstnode==0) {
			bstcomparison=0;
			while(1) {
				if(inbuf[inbufpos-bstcomparison]>
					inbuf[bstparent-bstcomparison]) {
					bstrightchild[bstparent]=inbufpos;
					break;
				}
				if(inbuf[inbufpos-bstcomparison]<
					inbuf[bstparent-bstcomparison]) {
					bstleftchild[bstparent]=inbufpos;
					break;
				}
				bstcomparison++;
				if(bstcomparison>bstlimit) {
					break;
				}
			}
			break;
		} else {
			if(bstnode+1<(UINT32)(inbufprefix+inbufsize))
				tmppredictedsymbol=inbuf[bstnode+1];
			bstcomparison=0;
			while(1) {
				if(inbuf[inbufpos-bstcomparison]>
					inbuf[bstnode-bstcomparison]) {
					bstparent=bstnode;
					bstnode=bstrightchild[bstnode];
					break;
				}
				if(inbuf[inbufpos-bstcomparison]<
					inbuf[bstnode-bstcomparison]) {
					bstparent=bstnode;
					bstnode=bstleftchild[bstnode];
					break;
				}
				bstcomparison++;
				if(bstcomparison>=bstmatchlen) {
					//---longest common suffix match
					bstmatchlen=bstcomparison;
					predictedsymbol=tmppredictedsymbol;
				}
				if(bstcomparison>bstlimit) {
					bstparent=bstnode;
					bstnode=0;
					break;
				}
			}
		}
	}
}

//------------------------------------------------------------------------------
// UPDATE RANKTABLE
//------------------------------------------------------------------------------
inline void updateranktable(UINT16 rctx, UINT8 j)
{
	UINT32 i;
	UINT32 tmp1,tmp2;
	UINT8 incrementvalue=0;

	//---increment counter
	incrementvalue=((255-j)>>2)+1;
	ranktable[rctx][j].count+=incrementvalue;

	//---scale counts down
	if(ranktable[rctx][j].count>65471) {
		for(i=0;i<NRANKS;i++) {
			ranktable[rctx][i].count=
				(UINT16)(ranktable[rctx][i].count>>1);
			if(ranktable[rctx][i].count==0)
				ranktable[rctx][i].count++;
		}
	}

	//---keep table sorted by descendending counts
	while (j>0 && ranktable[rctx][j].count>=ranktable[rctx][j-1].count) {
		tmp1=ranktable[rctx][j].count;
		tmp2=ranktable[rctx][j].character;
		ranktable[rctx][j].count=ranktable[rctx][j-1].count;
		ranktable[rctx][j].character=ranktable[rctx][j-1].character;
		ranktable[rctx][j-1].count=tmp1;
		ranktable[rctx][j-1].character=tmp2;
		j--;
	}
}

//------------------------------------------------------------------------------
// INIT MEMORY
//------------------------------------------------------------------------------
inline void initmemory()
{
	UINT32 i,ii;

	switch (buffermode) {
		case 0: inbufsize=1*1024*512;break;
		case 1: inbufsize=1*1024*1024;break;
		case 2: inbufsize=2*1024*1024;break;
		case 3: inbufsize=4*1024*1024;break;
		case 4: inbufsize=8*1024*1024;break;
		case 5: inbufsize=16*1024*1024;break;
		case 6: inbufsize=32*1024*1024;break;
		case 7: inbufsize=64*1024*1024;break;
		case 8: inbufsize=128*1024*1024;break;
		case 9: inbufsize=256*1024*1024;break;
	}

	switch (suffixmode) {
		case 0: sfxlen=1;break;
		case 1: sfxlen=1;break;
		case 2: sfxlen=2;break;
		case 3: sfxlen=4;break;
		case 4: sfxlen=8;break;
		case 5: sfxlen=16;break;
		case 6: sfxlen=32;break;
		case 7: sfxlen=64;break;
		case 8: sfxlen=128;break;
		case 9: sfxlen=256;break;
	}
	bstlimit=sfxlen;
	sfxlen+=1; //to add end-of-suffix marker

	switch (tstmemmode) {
		case 0: tstmemsize=7*1*1024*512;break;
		case 1: tstmemsize=7*1*1024*1024;break;
		case 2: tstmemsize=7*2*1024*1024;break;
		case 3: tstmemsize=7*4*1024*1024;break;
		case 4: tstmemsize=7*8*1024*1024;break;
		case 5: tstmemsize=7*16*1024*1024;break;
		case 6: tstmemsize=7*32*1024*1024;break;
		case 7: tstmemsize=7*64*1024*1024;break;
		case 8: tstmemsize=7*128*1024*1024;break;
		case 9: tstmemsize=7*256*1024*1024;break;
	}

	switch (contextmode) {
		case 0:
			//16bit context
			ctxmem=0x0000FFFF;
			mask=0x0000FFFF;
			break;
		case 1:
			//20bit context
			ctxmem=0x000FFFFF;
			mask=0x000FFFFF;
			break;
		case 2:
			//24bit context
			ctxmem=0x00FFFFFF;
			mask=0x00FFFFFF;
			break;
	}

	//---init ranktable
	for(ii=0;ii<NRANKCTX;ii++) {
		for(i=0;i<NRANKS;i++) {
			ranktable[ii][i].character=(UINT8)i;
			ranktable[ii][i].count=1;
		}
	}

	//---init input buffer
	inbuf=(UINT8 *)malloc(sizeof(UINT8)*(inbufsize+inbufprefix));
	if (inbuf==NULL)quit("out of memory");
	for(i=0;i<inbufsize+inbufprefix;i++) {
		inbuf[i]=i%256;
	}

	if(searchmode){
		//---init binary search
		bstrightchild=(UINT32 *)malloc(sizeof(UINT32)*(inbufsize+inbufprefix));
		if (bstrightchild==NULL)quit("out of memory");
		bstleftchild=(UINT32 *)malloc(sizeof(UINT32)*(inbufsize+inbufprefix));
		if (bstleftchild==NULL)quit("out of memory");
		for(i=0;i<inbufsize+inbufprefix;i++) {
			bstrightchild[i]=0;
			bstleftchild[i]=0;
		}
	}else{
		//---init suffix
		suffix=(UINT16 *)malloc(sizeof(UINT16)*(sfxlen));
		if (suffix==NULL)quit("out of memory");
	
		//---enough memory for TST left ?
		tstmemtest=(UINT8 *)malloc(sizeof(UINT8)*tstmemsize);
		if (tstmemtest==NULL)quit("out of memory");
		free(tstmemtest);
	
		//---init suffix tree
		root=tstgetroot();
		tstmaxnodes=tstmemsize/sizeof(Tnode);
		tstmaxnodes-=sfxlen;
	}

	if(verbosemode){

		if(searchmode){
			printf("Search mode: BST (binary search tree)\n");
		}else{
			printf("Search mode: TST (ternary search tree)\n");
		}

		printf("Buffer size: %d bytes\n",inbufsize);
		printf("Suffix length limit: %d bytes\n",sfxlen-1);

		if(searchmode){
			printf("BST memory usage: %d bytes\n",
				(inbufsize+inbufprefix)*2*sizeof(UINT32));
		}else{
			printf("TST memory limit: %d bytes (= %d nodes)\n",
				tstmemsize,tstmaxnodes);
			if(lazymode){
				printf("TST lazy node creation mode: on\n");
			}else{
				printf("TST lazy node creation mode: off\n");
			}
		}

		printf("using %d bytes memory for rank table\n",
			NRANKCTX*NRANKS*sizeof(ranktype));
		printf("using %d bytes memory for secondary contexts\n",
			(ctxmem+1)*4);
	}
}

//------------------------------------------------------------------------------
// WRITE HEADER
//------------------------------------------------------------------------------
inline void writeheader(FILE *out)
{
	fputc(MAGICBYTE,out);
	fputc(VERSION,out);
	fputc(SUBVERSION,out);
	fputc(buffermode,out);
	fputc(suffixmode,out);
	fputc(tstmemmode,out);
	fputc(rebuildmode,out);
	fputc(lazymode,out);
	fputc(contextmode,out);
	fputc(ctxconstructionmode,out);
	fputc(searchmode,out);
	headerlength=11;
}

//------------------------------------------------------------------------------
// READ HEADER
//------------------------------------------------------------------------------
inline SINT32 readheader(FILE *in)
{
	if(fgetc(in)!=MAGICBYTE)return 1;
	if(fgetc(in)!=VERSION)return 1;
	if(fgetc(in)!=SUBVERSION)return 1;
	buffermode=fgetc(in);if(buffermode>9)return 1;
	suffixmode=fgetc(in);if(suffixmode>9)return 1;
	tstmemmode=fgetc(in);if(tstmemmode>9)return 1;
	rebuildmode=fgetc(in);if(rebuildmode>9)return 1;
	lazymode=fgetc(in);if(lazymode>1)return 1;
	contextmode=fgetc(in);if(contextmode>2)return 1;
	ctxconstructionmode=fgetc(in);if(ctxconstructionmode>1)return 1;
	searchmode=fgetc(in);if(searchmode>1)return 1;
	headerlength=11;
	return 0;
}

//------------------------------------------------------------------------------
// ENCODE
//------------------------------------------------------------------------------
SINT32 encode(char *infile,char *outfile)
{
	clock_t begintime = clock();
	UINT32 endoffile=(UINT32)EOF;
	SINT32 i=0,j=0;

	//---init memory
	initmemory();

	//---open files
	in=fopen(infile,"rb");
	out=fopen(outfile,"wb");
	if(in==NULL || out==NULL) {
		fclose(in);
		fclose(out);
		printf("can't open files\n");
		return 1;
	}

	//---write header
	writeheader(out);

	//---process input file
	Encoder e(out,ctxmem+1);
	while(ch=(UINT32)fgetc(in),ch!=endoffile) {
		inbytes++;

		//---get rank
		for(i=0;i<(SINT32)NRANKS;i++) {
			if(ranktable[rkctx][i].character==(UINT8)ch) {
				j=i;
				break;
			}
		}

		//---encode rank
		switch (ctxconstructionmode){
			case 0:
				pctx=((ctx<<6)+(predictedsymbol&0x3F))&mask;
				for (i=0;i<j;++i){
					e.code(pctx,1);
					pctx<<=1;
					pctx+=1;
					pctx=pctx&mask;
				}
				e.code(pctx,0);
				break;
			case 1:
				for (i=0;i<j;++i){
					pctx=((i<<6)+(predictedsymbol&0x3F))&mask;
					e.code(pctx,1);
				}
				pctx=((j<<6)+(predictedsymbol&0x3F))&mask;
				e.code(pctx,0);
				break;
		}

		//---update ranktable,
		//---binary or ternary search tree and contexts
		updateranktable(rkctx,j);
		if(searchmode){
			updatebst(ch);
		}else{
			updatetst(ch);
		}
		ctx=((ctx<<6)+(ch&0x3F))&mask;
		rkctx=((ctx<<8)+predictedsymbol)&0x0FFF;

		if(verbosemode){
			//---display result
			if(inbytes%0xFFFF==0)
				printf("%d bytes in %d bytes out ratio: %f\n",
				inbytes,outbytes,(float)outbytes/inbytes);
		}
	}//eof

	//---EOF coded as rank 256
	j=256;
	switch (ctxconstructionmode){
		case 0:
			pctx=((ctx<<6)+(predictedsymbol&0x3F))&mask;
			for (i=0;i<j;++i){
				e.code(pctx,1);
				pctx<<=1;
				pctx+=1;
				pctx=pctx&mask;
			}
			e.code(pctx,0);
			break;
		case 1:
			for (i=0;i<j;++i){
				pctx=((i<<6)+(predictedsymbol&0x3F))&mask;
				e.code(pctx,1);
			}
			pctx=((j<<6)+(predictedsymbol&0x3F))&mask;
			e.code(pctx,0);
			break;
	}

	//---flush
	e.flush();

	//---close Files
	fclose(in);
	fclose(out);

	//---add header length + correction
	outbytes+=headerlength+1;

	//---display result
	printf ("%d Bytes encoded to %d Bytes\n",inbytes,outbytes);
	printf ("Ratio: %f = %f bpb\n",
		(float)outbytes/inbytes,((float)outbytes/inbytes)*8);
	totaltime=((double)(clock()-begintime))/CLOCKS_PER_SEC;
	printf("Time : %.3f sec\n",totaltime);
	printf("Speed: %.3f Mbyte/s in, %.3f Mbyte/s out\n",
		(inbytes/totaltime)/(1024*1024),
		(outbytes/totaltime)/(1024*1024));

	//---end of encoding
	return 0;
}

//------------------------------------------------------------------------------
// DECODE
//------------------------------------------------------------------------------
SINT32 decode(char *infile,char *outfile)
{
	clock_t begintime = clock();
	SINT32 j=0,cont=0;

	//---open files
	in=fopen(infile,"rb");
	out=fopen(outfile,"wb");
	if(in==NULL || out==NULL) {
		fclose(in);
		fclose(out);
		printf("can't open files\n");
		return 1;
	}

	//---read header
	if(readheader(in)==1){
		printf("unknown input file format\n");
		return 1;
	}

	//---init memory
	initmemory();

	//---decode rank
	Decoder d(in,ctxmem+1);
	switch (ctxconstructionmode){
		case 0:
			while (1) {
				pctx=((ctx<<6)+(predictedsymbol&0x3F))&mask;
				j=0;
				while(d.code(pctx)){
					j++;
					pctx<<=1;
					pctx+=1;
					pctx=pctx&mask;
				}

				if(j==256)break; //EOF?
		
				ch=ranktable[rkctx][j].character;
				fputc(ch,out);
				outbytes++;
		
				//---update ranktable,
				//---binary or ternary search tree and contexts
				updateranktable(rkctx,j);
				if(searchmode){
					updatebst(ch);
				}else{
					updatetst(ch);
				}
				ctx=((ctx<<6)+(ch&0x3F))&mask;
				rkctx=((ctx<<8)+predictedsymbol)&0x0FFF;
			}
			break;
		case 1:
			while (1) {
				j=0;
				cont=1;
				while(cont){
					pctx=((j<<6)+(predictedsymbol&0x3F))&mask;
					cont=d.code(pctx);
					j++;
				}
				j--;

				if(j==256)break; //EOF?
		
				ch=ranktable[rkctx][j].character;
				fputc(ch,out);
				outbytes++;

				//---update ranktable,
				//---binary or ternary search tree and contexts
				updateranktable(rkctx,j);
				if(searchmode){
					updatebst(ch);
				}else{
					updatetst(ch);
				}
				ctx=((ctx<<6)+(ch&0x3F))&mask;
				rkctx=((ctx<<8)+predictedsymbol)&0x0FFF;
			}
			break;
	}

	//---close Files
	fclose(in);
	fclose(out);

	//---add header bytes - correction
	inbytes+=headerlength-3;

	//---display result
	printf("%d Bytes decoded to %d Bytes\n",inbytes,outbytes);
	printf("Ratio: %f = %f bpb\n",
		(float)outbytes/inbytes,((float)outbytes/inbytes)*8);
	totaltime=((double)(clock()-begintime))/CLOCKS_PER_SEC;
	printf("Time : %.3f sec\n",totaltime);
	printf("Speed: %.3f Mbyte/s in, %.3f Mbyte/s out\n",
		(inbytes/totaltime)/(1024*1024),
		(outbytes/totaltime)/(1024*1024));

	//---end of decoding
	return 0;
}

//------------------------------------------------------------------------------
// USAGE
//------------------------------------------------------------------------------
void usage()
{
	printf("Usage: contra [options] <infile> <outfile>");
	printf("\nOptions are:");
	printf("\n-s[0-1] Suffix match search mode: (0=ternary tree, 1=binary tree) (default=0)");
	printf("\n-b[0-9] Buffer size mode (default=0) (binary tree uses 8x buffer memory)");
	printf("\n            mode / buffer sizes in megabytes are:");
	printf("\n            0=0.5, 1=1, 2=2, 3=4, 4=8, 5=16, 6=32, 7=64, 8=128, 9=256");
	printf("\n-r[0-9] Rebuild mode (default=3)");
	printf("\n            mode / tree rebuilds from last x percent of buffer:");
	printf("\n            0=0, 1=10, 2=20, 3=30, 4=40, 5=50, 6=60, 7=70, 8=80, 9=90");
	printf("\n-t[0-9] Ternary search tree: memory usage limit mode (default=0)");
	printf("\n            mode / tree memory usage limits in megabytes are:");
	printf("\n            0=3.5, 1=7, 2=14, 3=28, 4=56, 5=112, 6=224, 7=448, 8=896, 9=1792");
	printf("\n-l[0-9] Ternary search tree: suffix length limit mode (default=3)");
	printf("\n            mode / suffix length limits in bytes are:");
	printf("\n            0=1, 1=1, 2=2, 3=4, 4=8, 5=16, 6=32, 7=64, 8=128, 9=256");
	printf("\n-n[0-1] Ternary search tree: lazy node creation mode (0=off,1=on) (default=0)");
	printf("\n-m[0-2] Secondary context: memory usage mode (0=256KB 1=4MB 2=64MB) (default=0)");
	printf("\n-c[0-1] Secondary context: construction mode (0 or 1) (default=0)");
	printf("\n-v[0-1] Verbose mode (0=off, 1=on) (default=0)");
	printf("\n-d      Decodes <infile> to <outfile>");
	printf("\nThis program is free software; you can redistribute it and/or modify it under");
	printf("\nthe terms of the GNU General Public License. Files with a size >2GB are not");
	printf("\nsupported. Please report bugs to: frank.schwellinger@gmx.de\n");
}

//------------------------------------------------------------------------------
// MAIN
//------------------------------------------------------------------------------
SINT32 main(SINT32 argc,char *argv[] )
{
	SINT32 i;
	SINT32 codecmode,innum=0,outnum=0;
	FILE *in,*out;
	printf("contra %d.%d - a context based symbol ranking file (de-)compressor\n",
		VERSION,SUBVERSION);
	//---default values
	codecmode=0;
	buffermode=0;
	suffixmode=3;
	tstmemmode=0;
	rebuildmode=3;
	lazymode=0;
	contextmode=0;
	verbosemode=0;
	ctxconstructionmode=0;
	searchmode=0;
	//---scan command line
	if(argc<3) {
		usage();
		exit(1);
	}
	for(i=1;i<argc;i++) {
		char *sz=argv[i];
		if(*sz=='-') {
			switch(*(sz+1)) {
				case 'm':
				case 'M':
					//---secondary context memory usage mode
					switch(*(sz+2)) {
						case '0':
							contextmode=0;
							break;
						case '1':
							contextmode=1;
							break;
						case '2':
							contextmode=2;
							break;
						default:
							printf("bad argument: allowed is secondary context memory usage mode 0-2\n");
							exit(1);
					}
					break;
				case 'c':
				case 'C':
					//---Secondary context construction mode
					switch(*(sz+2)) {
						case '0':
							ctxconstructionmode=0;
							break;
						case '1':
							ctxconstructionmode=1;
							break;
						default:
							printf("bad argument: allowed is secondary context construction mode 0-1\n");
							exit(1);
					}
					break;
				case 'b':
				case 'B':
					//---Buffer mode
					switch(*(sz+2)) {
						case '0':
							buffermode=0;
							break;
						case '1':
							buffermode=1;
							break;
						case '2':
							buffermode=2;
							break;
						case '3':
							buffermode=3;
							break;
						case '4':
							buffermode=4;
							break;
						case '5':
							buffermode=5;
							break;
						case '6':
							buffermode=6;
							break;
						case '7':
							buffermode=7;
							break;
						case '8':
							buffermode=8;
							break;
						case '9':
							buffermode=9;
							break;
						default:
							printf("bad argument: allowed is buffer mode 0-9\n");
							exit(1);
					}
					break;
				case 'l':
				case 'L':
					//---Suffix length limit mode
					switch(*(sz+2)) {
						case '0':
							suffixmode=0;
							break;
						case '1':
							suffixmode=1;
							break;
						case '2':
							suffixmode=2;
							break;
						case '3':
							suffixmode=3;
							break;
						case '4':
							suffixmode=4;
							break;
						case '5':
							suffixmode=5;
							break;
						case '6':
							suffixmode=6;
							break;
						case '7':
							suffixmode=7;
							break;
						case '8':
							suffixmode=8;
							break;
						case '9':
							suffixmode=9;
							break;
						default:
							printf("bad argument: allowed is suffix length limit mode 0-9\n");
							exit(1);
					}
					break;
				case 't':
				case 'T':
					//---TST memory usage mode
					switch(*(sz+2)) {
						case '0':
							tstmemmode=0;
							break;
						case '1':
							tstmemmode=1;
							break;
						case '2':
							tstmemmode=2;
							break;
						case '3':
							tstmemmode=3;
							break;
						case '4':
							tstmemmode=4;
							break;
						case '5':
							tstmemmode=5;
							break;
						case '6':
							tstmemmode=6;
							break;
						case '7':
							tstmemmode=7;
							break;
						case '8':
							tstmemmode=8;
							break;
						case '9':
							tstmemmode=9;
							break;
						default:
							printf("bad argument: allowed is ternary search tree memory usage mode 0-9\n");
							exit(1);
					}
					break;
				case 'r':
				case 'R':
					//---rebuild mode
					switch(*(sz+2)) {
						case '0':
							rebuildmode=0;
							break;
						case '1':
							rebuildmode=1;
							break;
						case '2':
							rebuildmode=2;
							break;
						case '3':
							rebuildmode=3;
							break;
						case '4':
							rebuildmode=4;
							break;
						case '5':
							rebuildmode=5;
							break;
						case '6':
							rebuildmode=6;
							break;
						case '7':
							rebuildmode=7;
							break;
						case '8':
							rebuildmode=8;
							break;
						case '9':
							rebuildmode=9;
							break;
						default:
							printf("bad argument: allowed is rebuild mode 0-9\n");
							exit(1);
					}
					break;
				case 'n':
				case 'N':
					//---lazy node creation mode
					switch(*(sz+2)) {
						case '0':
							lazymode=0;
							break;
						case '1':
							lazymode=1;
							break;
						default:
							printf("bad argument: allowed is lazy node creation mode 0-1\n");
							exit(1);
					}
					break;
				case 's':
				case 'S':
					//---Search tree mode
					switch(*(sz+2)) {
						case '0':
							searchmode=0; //ternary search tree
							break;
						case '1':
							searchmode=1; //binary search tree
							break;
						default:
							printf("bad argument: allowed is suffix match search mode 0-1\n");
							exit(1);
					}
					break;
				case 'v':
				case 'V':
					//---verbose mode
					switch(*(sz+2)) {
						case '0':
							verbosemode=0;
							break;
						case '1':
							verbosemode=1;
							break;
						default:
							printf("bad argument: allowed is verbose mode 0-1\n");
							exit(1);
					}
					break;
				case 'd':
				case 'D':
					//---decode
					if(argc!=4) {
						usage();
						exit(1);
					} else {
						codecmode=1;
						break;
					}
				default:
					printf("unknown argument\n");
					exit(1);
			}
		} else {
			if(argv[i+2]!=NULL) {
				printf("bad arguments\n");
				exit(1);
			} else {
				innum=i;
				outnum=i+1;
				break;
			}
		}
	}
	//---test input file
	in=fopen(argv[innum],"rb");
	if(in==NULL) {
		printf("can't open input file\n");
		exit(1);
	}
	fclose(in);
	//---test output file
	out=fopen(argv[outnum],"wb");
	if(out==NULL) {
		printf("can't open output file\n");
		exit(1);
	}
	fclose(out);
	//---encode/decode
	if(codecmode==0)
		return encode(argv[innum],argv[outnum]);
	if(codecmode==1)
		return decode(argv[innum],argv[outnum]);
}

//------------------------------------------------------------------------------
// END OF PROGRAM
//------------------------------------------------------------------------------
